/*
 * "Memory" driver
 * Copyright (c) 2010 Michal Nazarewicz (mina86@mina86.com)
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/uaccess.h>

MODULE_DESCRIPTION("Example 'memory' driver");
MODULE_AUTHOR("Michal Nazarewicz");
MODULE_LICENSE("GPL");


static char memory[PAGE_SIZE];


static loff_t mem_llseek(struct file *file, loff_t off, int whence)
{
	switch (whence) {
	case SEEK_CUR:
		off += file->f_pos;
		break;
	case SEEK_END:
		off += sizeof memory;
	}
	if (off < 0)
		return -EINVAL;
	file->f_pos = off;
	return off;
}


static ssize_t mem_read(struct file *file, char __user *buf,
						 size_t len, loff_t *offp)
{
	printk(KERN_INFO "mem: read(%zu)\n", len);

	if (*offp >= sizeof memory || !len)
		return 0;

	len = min(len, (size_t)(sizeof memory - *offp));
	if (unlikely(__copy_to_user(buf, memory + *offp, len)))
		return -EFAULT;

	*offp += len;
	return len;
}


static ssize_t mem_write(struct file *file, const char __user *buf,
						  size_t len, loff_t *offp)
{
	printk(KERN_INFO "mem: write(%zu)\n", len);

	if (!len)
		return 0;

	if (*offp >= sizeof memory)
		return -ENOSPC;

	len = min(len, (size_t)(sizeof memory - *offp));
	if (unlikely(__copy_from_user(memory + *offp, buf, len)))
		return -EFAULT;

	*offp += len;
	return len;
}


static const struct file_operations mem_fops = {
	.llseek  = mem_llseek,
	.read    = mem_read,
	.write   = mem_write,
};

static struct miscdevice mem_dev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name  = "mymem",
	.fops  = &mem_fops,
};


static int __init mem_init(void)
{
	return misc_register(&mem_dev);
}
module_init(mem_init)

static void __exit mem_exit(void)
{
	misc_deregister(&mem_dev);
}
module_exit(mem_exit)
