/*
 * "Memory" driver, growable
 * Copyright (c) 2010 Michal Nazarewicz (mina86@mina86.com)
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/vmalloc.h>

MODULE_DESCRIPTION("Example growable 'memory' driver");
MODULE_AUTHOR("Michal Nazarewicz");
MODULE_LICENSE("GPL");


#define mem_max_size (size_t)(1 << 20)
static size_t mem_capacity;
static size_t mem_size;
static char *memory;


static int ensure_capacity(size_t capacity)
{
	capacity = (capacity + ~PAGE_MASK) & PAGE_MASK;

	if (mem_capacity < capacity) {
		char *tmp = vmalloc(capacity);
		if (unlikely(!tmp))
			return -ENOMEM;

		printk(KERN_INFO "mem: new capacity %zu\n", capacity);

		if (memory) {
			memcpy(tmp, memory, mem_size);
			vfree(memory);
		}

		memory = tmp;
		mem_capacity = capacity;
	}

	return 0;
}


static loff_t mem_llseek(struct file *file, loff_t off, int whence)
{
	switch (whence) {
	case SEEK_CUR:
		off += file->f_pos;
		break;
	case SEEK_END:
		off += mem_size;
	}
	if (off < 0)
		return -EINVAL;
	file->f_pos = off;
	return off;
}


static ssize_t mem_read(struct file *file, char __user *buf,
						 size_t len, loff_t *offp)
{
	printk(KERN_INFO "mem: read(%zu)\n", len);

	if (*offp >= mem_size || !len)
		return 0;

	len = min(len, (size_t)(mem_size - *offp));
	if (unlikely(__copy_to_user(buf, memory + *offp, len)))
		return -EFAULT;

	*offp += len;
	return len;
}


static ssize_t mem_write(struct file *file, const char __user *buf,
						  size_t len, loff_t *offp)
{
	int ret;

	printk(KERN_INFO "mem: write(%zu)\n", len);

	if (!len)
		return 0;

	if (*offp >= mem_max_size)
		return -ENOSPC;

	ret = ensure_capacity(min((size_t)(*offp+len), mem_max_size));
	if (unlikely(ret < 0))
		return ret;

	len = min(len, (size_t)(mem_capacity - *offp));
	if (unlikely(__copy_from_user(memory + *offp, buf, len)))
		return -EFAULT;

	*offp += len;
	mem_size = max((size_t)*offp, mem_size);
	printk(KERN_INFO "mem: new size %zu\n", mem_size);
	return len;
}


static const struct file_operations mem_fops = {
	.llseek  = mem_llseek,
	.read    = mem_read,
	.write   = mem_write,
};

static struct miscdevice mem_dev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name  = "mymem",
	.fops  = &mem_fops,
};


static int __init mem_init(void)
{
	return misc_register(&mem_dev);
}
module_init(mem_init)

static void __exit mem_exit(void)
{
	misc_deregister(&mem_dev);
	vfree(memory);
}
module_exit(mem_exit)
