/*
 * "Memory" driver, growable with synchronization
 * Copyright (c) 2010 Michal Nazarewicz (mina86@mina86.com)
 */

#include <linux/module.h>
#include <linux/miscdevice.h>
#include <linux/fs.h>
#include <linux/uaccess.h>
#include <linux/vmalloc.h>
#include <linux/mutex.h>

MODULE_DESCRIPTION("Example growable 'memory' driver with synchronisation");
MODULE_AUTHOR("Michal Nazarewicz");
MODULE_LICENSE("GPL");


static DEFINE_MUTEX(mem_mutex);

#define mem_max_size (size_t)(1 << 20)    /* mult. of PAGE_SIZE */
static size_t mem_capacity;
static size_t mem_size;
static char *memory;                      /* guarded by mutex */


static int __ensure_capacity(size_t capacity)
{
	capacity = (capacity + ~PAGE_MASK) & PAGE_MASK;

	if (mem_capacity < capacity) {
		char *tmp = vmalloc(capacity);
		if (unlikely(!tmp))
			return -ENOMEM;

		printk(KERN_INFO "mem: new capacity %zu\n", capacity);

		if (memory) {
			memcpy(tmp, memory, mem_size);
			vfree(memory);
		}

		memory = tmp;
		mem_capacity = capacity;
	}

	return 0;
}


static loff_t mem_llseek(struct file *file, loff_t off, int whence)
{
	switch (whence) {
	case SEEK_CUR:
		off += file->f_pos;
		break;
	case SEEK_END:
		off += mem_size;
	}
	if (off < 0)
		return -EINVAL;
	file->f_pos = off;
	return off;
}


static ssize_t mem_read(struct file *file, char __user *buf,
						 size_t len, loff_t *offp)
{
	ssize_t ret;

	printk(KERN_INFO "mem: read(%zu)\n", len);

	if (*offp >= mem_size || !len)
		return 0;

	if (unlikely(mutex_lock_interruptible(&mem_mutex)))
		return -EINTR;

	len = min(len, (size_t)(mem_size - *offp));
	ret = len;
	if (unlikely(__copy_to_user(buf, memory + *offp, len)))
		ret = -EFAULT;
	else
		*offp += len;

	mutex_unlock(&mem_mutex);
	return ret;
}


static ssize_t mem_write(struct file *file, const char __user *buf,
						  size_t len, loff_t *offp)
{
	ssize_t ret;

	printk(KERN_INFO "mem: write(%zu)\n", len);

	if (!len)
		return 0;

	if (*offp >= mem_max_size)
		return -ENOSPC;

	if (unlikely(mutex_lock_interruptible(&mem_mutex)))
		return -EINTR;

	ret = __ensure_capacity(min((size_t)(*offp + len), (size_t)mem_max_size));
	if (unlikely(ret < 0))
		goto done;

	len = min(len, (size_t)(mem_capacity - *offp));
	ret = -EFAULT;
	if (likely(!__copy_from_user(memory + *offp, buf, len))) {
		ret = len;
		*offp += len;
		mem_size = max((size_t)*offp, mem_size);
		printk(KERN_INFO "mem: new size %zu\n", mem_size);
	}

done:
	mutex_unlock(&mem_mutex);
	return ret;
}


static const struct file_operations mem_fops = {
	.llseek  = mem_llseek,
	.read    = mem_read,
	.write   = mem_write,
};

static struct miscdevice mem_dev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name  = "mymem",
	.fops  = &mem_fops,
};


static int __init mem_init(void)
{
	return misc_register(&mem_dev);
}
module_init(mem_init)

static void __exit mem_exit(void)
{
	misc_deregister(&mem_dev);
	vfree(memory);
}
module_exit(mem_exit)
